@@ -1,5 +1,24 @@
 #!/usr/bin/env python
 
+# Copyright (C) 2020-2021  Asitha Senanayake
+#
+# This file is part of python_ags4.
+#
+# python_ags4 is free software: you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program.  If not, see <https://www.gnu.org/licenses/>.
+#
+# https://github.com/asitha-sena/python-ags4
+
 import click
 from python_ags4 import AGS4, __version__
 from rich.console import Console
@@ -110,38 +129,42 @@ def check(input_file, dictionary, output_file):
 
         ags_errors = AGS4.check_file(input_file, standard_AGS4_dictionary=dictionary)
 
-        # Dictionay evaluates to False if empty
-        if bool(ags_errors) is False:
-            console.print('\n[green]File check complete! No errors found.[/green]\n')
+        # Count number of entries in error log
+        error_count = 0
+        for key, val in ags_errors.items():
+            if 'Rule' in key:
+                error_count += len(val)
+
+        # Print "All checks passed!" to screen if no errors are found
+        if error_count == 0:
+            print_to_screen(ags_errors)
+            console.print('\n[green]File check complete! All checks passed![/green]')
 
             if output_file is not None:
-                save_to_file(output_file, ags_errors, input_file, 'No')
+                save_to_file(output_file, ags_errors, input_file, error_count)
+                console.print(f'\n[green]Report saved in {output_file}[/green]\n')
+
+        # Print errors to screen if list is short enough
+        elif error_count < 100:
+            print_to_screen(ags_errors)
+
+            console.print(f'\n[yellow]File check complete! {error_count} errors found![/yellow]')
+
+            if output_file is not None:
+                save_to_file(output_file, ags_errors, input_file, error_count)
+                console.print(f'\n[yellow]Error report saved in {output_file}[/yellow]\n')
 
         else:
-            # Count number of entries in error log
-            error_count = 0
-            for key, val in ags_errors.items():
-                if 'Rule' in key:
-                    error_count += len(val)
+            console.print(f'\n[yellow]File check complete! {error_count} errors found![/yellow]')
+            console.print('\n[yellow]Error report too long to print to screen.[/yellow]')
 
-            # Print errors to screen if list is short enough.
-            if error_count < 100:
-                print_to_screen(ags_errors)
+            # Assign default path if output_file is not provided
+            if output_file is None:
+                output_dir = os.path.dirname(input_file)
+                output_file = os.path.join(output_dir, 'error_log.txt')
 
-                console.print(f'\n[yellow]File check complete! {error_count} errors found![/yellow]')
-
-                if output_file is not None:
-                    save_to_file(output_file, ags_errors, input_file, error_count)
-
-            else:
-                console.print(f'\n[yellow]File check complete! {error_count} errors found![/yellow]')
-                console.print('\n[yellow]Error report too long to print to screen.[/yellow]')
-
-                if output_file is None:
-                    output_dir = os.path.dirname(input_file)
-                    output_file = os.path.join(output_dir, 'error_log.txt')
-
-                save_to_file(output_file, ags_errors, input_file, error_count)
+            save_to_file(output_file, ags_errors, input_file, error_count)
+            console.print(f'\n[yellow]Error report saved in {output_file}[/yellow]\n')
 
     else:
         console.print('[red]ERROR: Only .ags files are accepted as input.[/red]')
@@ -185,6 +208,13 @@ def save_to_file(output_file, ags_errors, input_file, error_count):
                     f.write(f'''{entry['line']}: \t {entry['desc']}\n''')
                 f.write('\n')
 
+            # Summary of errors log
+            if error_count == 0:
+                f.write('All checks passed!\n')
+            else:
+                f.write(f'{error_count} error(s) found in file!\n')
+                f.write('\n')
+
             # Write 'General' error messages first if present
             if 'General' in ags_errors.keys():
                 f.write('General:\n')
@@ -199,11 +229,6 @@ def save_to_file(output_file, ags_errors, input_file, error_count):
                     f.write(f'''  Line {entry['line']}\t {entry['group'].strip('"')}\t {entry['desc']}\n''')
                 f.write('\n')
 
-            if error_count == 'No':
-                console.print(f'[green]Report saved in {output_file}[/green]\n')
-            else:
-                console.print(f'\n[yellow]Error report saved in {output_file}[/yellow]\n')
-
     except FileNotFoundError:
         console.print('[red]\nERROR: Invalid output file path. Error report could not be saved.[/red]')
         console.print('[red]       Please ensure that the specified directory exists.[/red]')
